import copy
from typing import Optional
from unittest.mock import Mock

import pytest
from dbt_jobs_as_code.client import DBTCloud, DBTCloudException
from dbt_jobs_as_code.cloud_yaml_mapping.validate_link import LinkableCheck, can_be_linked
from dbt_jobs_as_code.schemas.job import JobDefinition


@pytest.fixture
def mock_dbt_cloud():
    return Mock(spec=DBTCloud)


@pytest.fixture
def base_job_definition():
    return JobDefinition(
        id=None,
        linked_id=123,
        identifier=None,
        account_id=1,
        project_id=1,
        environment_id=1,
        name="Test Job",
        settings={},
        run_generate_sources=False,
        execute_steps=[],
        generate_docs=False,
        schedule={"cron": "0 14 * * 0,1,2,3,4,5,6"},
        triggers={},
    )


def test_cant_be_linked_no_id_in_yaml(mock_dbt_cloud, base_job_definition):
    base_job_definition.linked_id = None

    result = can_be_linked("test_job", base_job_definition, mock_dbt_cloud)

    assert result.can_be_linked is False
    assert "doesn't have an ID in YAML" in result.message
    assert result.linked_job is None


def test_cant_be_linked_job_not_exist(mock_dbt_cloud, base_job_definition):
    mock_dbt_cloud.get_job.side_effect = DBTCloudException("Job not found")

    result = can_be_linked("test_job", base_job_definition, mock_dbt_cloud)

    assert result.can_be_linked is False
    assert "doesn't exist in dbt Cloud" in result.message
    assert result.linked_job is None
    mock_dbt_cloud.get_job.assert_called_once_with(job_id=123)


def test_cant_be_linked_already_linked(mock_dbt_cloud, base_job_definition):
    cloud_job = copy.deepcopy(base_job_definition)
    cloud_job.identifier = "existing_identifier"

    mock_dbt_cloud.get_job.return_value = cloud_job

    result = can_be_linked("test_job", base_job_definition, mock_dbt_cloud)

    assert result.can_be_linked is False
    assert "already linked" in result.message
    assert result.linked_job is None
    mock_dbt_cloud.get_job.assert_called_once_with(job_id=123)


def test_can_be_linked_success(mock_dbt_cloud, base_job_definition):
    cloud_job = copy.deepcopy(base_job_definition)
    cloud_job.id = 123
    mock_dbt_cloud.get_job.return_value = cloud_job

    result = can_be_linked("test_job", base_job_definition, mock_dbt_cloud)

    assert result.can_be_linked is True
    assert result.message == ""
    assert result.linked_job == cloud_job
    mock_dbt_cloud.get_job.assert_called_once_with(job_id=123)



import yaml
from beartype.typing import Optional, Set, TextIO
from jinja2 import Environment, StrictUndefined, meta
from jinja2.exceptions import UndefinedError
from loguru import logger

from dbt_jobs_as_code.schemas.config import Config


class LoadingJobsYAMLError(Exception):
    pass


def load_job_configuration(config_file: TextIO, vars_file: Optional[TextIO]) -> Config:
    """Load the job configuration set in a YAML file into a Config object

    Can be a non-templated YAML or a templated one for which we need to replace Jinja values
    """
    if vars_file:
        config = _load_yaml_with_template(config_file, vars_file)
    else:
        config = _load_yaml_no_template(config_file)

    if not config["jobs"]:
        return Config(jobs={})

    date_config = [job.get("schedule", {}).get("date", None) for job in config["jobs"].values()]
    time_config = [job.get("schedule", {}).get("time", None) for job in config["jobs"].values()]

    if any(date_config):
        logger.warning(
            f"⚡️ There is some date config under 'schedule > date' in your YML. This data is auto generated and should be deleted. Only cron is supported in the config."
        )
    if any(time_config):
        logger.warning(
            f"⚡️ There is some time config under 'schedule > time' in your YML. This data is auto generated and should be deleted. Only cron is supported in the config."
        )

    for identifier, job in config.get("jobs", {}).items():
        job["identifier"] = identifier

    return Config(**config)


def _load_yaml_no_template(config_file: TextIO) -> dict:
    """Load a job YAML file into a Config object"""
    config_string = config_file.read()

    jinja_vars = _get_jinja_variables(config_string)
    if jinja_vars:
        raise LoadingJobsYAMLError(
            f"This is a templated YAML file. Please remove the variables {jinja_vars} or provide the variables values."
        )

    return yaml.safe_load(config_string)


def _load_yaml_with_template(config_file: TextIO, vars_file: TextIO) -> dict:
    """Load a job YAML file into a Config object"""
    template_vars_values = yaml.safe_load(vars_file)
    config_string_unrendered = config_file.read()

    env = Environment(undefined=StrictUndefined)
    template = env.from_string(config_string_unrendered)

    try:
        config_string_rendered = template.render(template_vars_values)
    except UndefinedError as e:
        print(f"Error: {e}")  # This will raise an error
        raise LoadingJobsYAMLError(f"Some variables didn't have a value: {e.message}.") from e

    return yaml.safe_load(config_string_rendered)


def _get_jinja_variables(input: str) -> Set[str]:
    """Get the variables from a Jinja template"""
    env = Environment()
    parsed_input = env.parse(input)
    return meta.find_undeclared_variables(parsed_input)



from beartype.typing import List, Optional, TextIO
from loguru import logger

from dbt_jobs_as_code.client import DBTCloud
from dbt_jobs_as_code.loader.load import load_job_configuration
from dbt_jobs_as_code.schemas.job import JobDefinition


def get_account_id(config_file: Optional[TextIO], account_id: Optional[int]) -> int:
    """Get account ID from either config file or direct input"""
    if account_id:
        return account_id
    elif config_file:
        defined_jobs = load_job_configuration(config_file, None).jobs.values()
        return list(defined_jobs)[0].account_id
    else:
        raise ValueError("Either config or account_id must be provided")


def check_job_fields(dbt_cloud: DBTCloud, job_ids: List[int]) -> None:
    """Check if there are new fields in job model"""
    if not job_ids:
        logger.error("We need to provide some job_id to test the import")
        return

    logger.info("Checking if there are new fields for jobs")
    dbt_cloud.get_job_missing_fields(job_id=job_ids[0])


def fetch_jobs(
    dbt_cloud: DBTCloud, job_ids: List[int], project_ids: List[int], environment_ids: List[int]
) -> List[JobDefinition]:
    """Fetch jobs from dbt Cloud based on provided filters"""
    logger.info("Getting the jobs definition from dbt Cloud")

    if job_ids and not (project_ids or environment_ids):
        # Get jobs one by one if only job_ids provided
        cloud_jobs_can_have_none = [dbt_cloud.get_job(job_id=id) for id in job_ids]
        return [job for job in cloud_jobs_can_have_none if job is not None]

    # Get all jobs and filter
    cloud_jobs = dbt_cloud.get_jobs(project_ids=project_ids, environment_ids=environment_ids)
    if job_ids:
        cloud_jobs = [job for job in cloud_jobs if job.id in job_ids]
    return cloud_jobs



from beartype.typing import Any, Literal, Optional
from pydantic import BaseModel, Field, model_validator
from typing_extensions import Self


class CustomEnvironmentVariable(BaseModel):
    name: str
    type: Literal["project", "environment", "job", "user"] = "job"
    value: Optional[str] = Field(default=None)
    display_value: Optional[str] = None
    job_definition_id: Optional[int] = None

    @model_validator(mode="after")
    def check_env_var(self) -> Self:
        if not self.name.startswith("DBT_"):
            raise ValueError("Key must have `DBT_` prefix.")
        if not self.name.isupper():
            raise ValueError("Key name must be SCREAMING_SNAKE_CASE")
        return self


class CustomEnvironmentVariablePayload(CustomEnvironmentVariable):
    """A dbt Cloud-serializable representation of a CustomEnvironmentVariables."""

    id: Optional[int] = None
    project_id: int
    account_id: int
    raw_value: Optional[str] = None
    value: Optional[str] = Field(default=None, exclude=True)

    def __init__(self, **data: Any):
        data["raw_value"] = data["value"] if "value" in data else data["display_value"]
        super().__init__(**data)



import pytest
from dbt_jobs_as_code.loader.load import LoadingJobsYAMLError, load_job_configuration

# all the different ways of defining a set of jobs still create the same Pydantic jobs

expected_config_dict = {
    "jobs": {
        "job1": {
            "account_id": 43791,
            "compare_changes_flags": "--select state:modified",
            "custom_environment_variables": [],
            "dbt_version": None,
            "deferring_environment_id": None,
            "deferring_job_definition_id": None,
            "description": "",
            "environment_id": 134459,
            "execute_steps": [
                "dbt run --select model1+",
                "dbt run --select model2+",
                "dbt compile",
            ],
            "execution": {"timeout_seconds": 0},
            "generate_docs": False,
            "id": None,
            "identifier": "job1",
            "job_completion_trigger_condition": None,
            "job_type": "scheduled",
            "linked_id": None,
            "name": "My Job 1 with a new name",
            "project_id": 176941,
            "run_generate_sources": True,
            "schedule": {
                "cron": "0 */2 * * *",
                "date": {"cron": "0 */2 * * *", "type": "custom_cron"},
                "time": {"interval": 1, "type": "every_hour"},
            },
            "settings": {"target_name": "production", "threads": 4},
            "state": 1,
            "triggers": {
                "git_provider_webhook": False,
                "github_webhook": False,
                "on_merge": False,
                "schedule": True,
            },
            "triggers_on_draft_pr": False,
            "run_compare_changes": False,
        },
        "job2": {
            "account_id": 43791,
            "compare_changes_flags": "--select state:modified",
            "custom_environment_variables": [
                {
                    "display_value": None,
                    "job_definition_id": None,
                    "name": "DBT_ENV1",
                    "type": "job",
                    "value": "My val",
                },
                {
                    "display_value": None,
                    "job_definition_id": None,
                    "name": "DBT_ENV2",
                    "type": "job",
                    "value": "My val2",
                },
            ],
            "dbt_version": None,
            "deferring_environment_id": None,
            "deferring_job_definition_id": None,
            "description": "",
            "environment_id": 134459,
            "execute_steps": ["dbt run-operation clone_all_production_schemas", "dbt compile"],
            "execution": {"timeout_seconds": 0},
            "generate_docs": True,
            "id": None,
            "identifier": "job2",
            "job_completion_trigger_condition": {
                "condition": {"job_id": 123, "project_id": 234, "statuses": [10, 20]}
            },
            "job_type": "other",
            "linked_id": None,
            "name": "CI/CD run",
            "project_id": 176941,
            "run_generate_sources": True,
            "schedule": {
                "cron": "0 * * * *",
                "date": {"cron": "0 * * * *", "type": "custom_cron"},
                "time": {"interval": 1, "type": "every_hour"},
            },
            "settings": {"target_name": "TEST", "threads": 4},
            "state": 1,
            "triggers": {
                "git_provider_webhook": False,
                "github_webhook": True,
                "on_merge": True,
                "schedule": False,
            },
            "triggers_on_draft_pr": True,
            "run_compare_changes": True,
        },
    }
}


def test_load_yml_no_anchor():
    """Test that loading configuration without YML anchors works as expected."""

    with open("tests/loader/jobs.yml") as file:
        loaded_config = load_job_configuration(file, None)

    assert loaded_config.model_dump() == expected_config_dict


def test_load_yml_anchors():
    """Test that loading configuration with YML anchors works as expected."""

    with open("tests/loader/jobs_with_anchors.yml") as file:
        loaded_config = load_job_configuration(file, None)

    assert loaded_config.model_dump() == expected_config_dict


def test_load_yml_templated():
    """Test that load_job_configuration works with templated YAML and variables."""

    with open("tests/loader/jobs_templated.yml") as file:
        with open("tests/loader/jobs_templated_vars.yml") as templated_file:
            loaded_config = load_job_configuration(file, templated_file)

    assert loaded_config.model_dump() == expected_config_dict


def test_error_load_yml_templated_missing_vars_parameter():
    """Test that load_job_configuration works with templated YAML and variables."""

    with pytest.raises(LoadingJobsYAMLError) as exc_info:
        with open("tests/loader/jobs_templated.yml") as file:
            with open("tests/loader/jobs_templated_vars_missing.yml") as templated_file:
                load_job_configuration(file, templated_file)

    # we check that the error messages contains the missing variables
    assert "'environment_id'" in str(exc_info.value)


def test_error_load_yml_templated_missing_specific_var():
    """Test that load_job_configuration works with templated YAML and variables."""

    with pytest.raises(LoadingJobsYAMLError) as exc_info:
        with open("tests/loader/jobs_templated.yml") as file:
            load_job_configuration(file, None)

    # we check that the error messages contains the missing variables
    assert "'environment_id'" in str(exc_info.value)
    assert "'project_id'" in str(exc_info.value)



import json

from beartype.typing import Any, Optional, Tuple
from deepdiff import DeepDiff
from loguru import logger

from dbt_jobs_as_code.schemas.custom_environment_variable import (
    CustomEnvironmentVariable,
    CustomEnvironmentVariablePayload,
)
from dbt_jobs_as_code.schemas.job import JobDefinition


def _get_mismatched_dict_entries(
    dict_source: dict[str, Any], dict_dest: dict[str, Any]
) -> dict[str, Any]:
    """Returns a dict with the mismatched entries between two dicts"""

    return DeepDiff(dict_source, dict_dest, ignore_order=True)


def _job_to_dict(job: JobDefinition):
    dict_vals = job.model_dump(
        exclude={
            "id",  # we want to exclude id because our YAML file will not have it
            "custom_environment_variables",  # TODO: Add this back in. Requires extra API calls.
            "linked_id",  # we want to exclude linked_id because dbt Cloud doesn't save it
        }
    )
    return dict_vals


def check_job_mapping_same(source_job: JobDefinition, dest_job: JobDefinition) -> bool:
    """Checks if the source and destination jobs are the same"""

    source_job_dict = _job_to_dict(source_job)
    dest_job_dict = _job_to_dict(dest_job)

    diffs = _get_mismatched_dict_entries(dest_job_dict, source_job_dict)

    if len(diffs) == 0:
        logger.success(f"✅ Job {source_job.identifier} is identical")
        return True
    else:
        # we can't json.dump types normally, so we provide a custom logic to just return the type name
        def type_serializer(obj):
            if isinstance(obj, type):
                return obj.__name__
            raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")

        logger.info(
            f"❌ Job {source_job.identifier} is different - Diff:\n{json.dumps(diffs, indent=2, default=type_serializer)}"
        )
        return False


def check_env_var_same(
    source_env_var: CustomEnvironmentVariable,
    dest_env_vars: dict[str, CustomEnvironmentVariablePayload],
) -> Tuple[bool, Optional[int]]:
    """Checks if the source env vars is the same in the destination env vars"""

    if source_env_var.name not in dest_env_vars:
        raise Exception(
            f"Custom environment variable {source_env_var.name} not found in dbt Cloud, "
            f"you need to create it first."
        )

    env_var_id = dest_env_vars[source_env_var.name].id

    if dest_env_vars[source_env_var.name].value == source_env_var.value:
        logger.debug(
            f"The env var {source_env_var.name} is already up to date for the job {source_env_var.job_definition_id}."
        )
        return (True, env_var_id)
    else:
        logger.info(f"❌ The env var overwrite for {source_env_var.name} is different")
        return (False, env_var_id)
